@@ -17,12 +17,46 @@
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */
 
+/** \file unit.c */
+
+/**
+ * \addtogroup PedUnit
+ * 
+ * \brief The PedUnit module provides a standard mechanism for describing and parsing
+ * locations within devices in human-friendly plain text. 
+ *
+ * Internally, libparted uses PedSector (which is typedef'ed to be long long 
+ * in <parted/device.h>) to describe device locations such as the start and 
+ * end of partitions.  However, sector numbers are often long and unintuitive.  
+ * For example, my extended partition starts at sector 208845.  PedUnit allows 
+ * this location to be represented in more intutitive ways, including "106Mb", 
+ * "0Gb" and "0%", as well as "208845s".  PedUnit aims to provide facilities 
+ * to provide a consistent system for describing device locations all throughout 
+ * libparted.
+ * 
+ * PedUnit provides two basic services: converting a PedSector into a text
+ * representation, and parsing a text representation into a PedSector.
+ * PedUnit currently supports these units:
+ * 
+ * 	sectors, bytes, kilobytes, megabytes, gigabytes, terabytes, compact,
+ * 	cylinder and percent.
+ * 	
+ * PedUnit has a global variable that contains the default unit for all
+ * conversions.
+ *
+ * @{
+ */
+
+
+
+
 #include <parted/parted.h>
 #include <parted/debug.h>
 
 #include <ctype.h>
 #include <stdio.h>
 #include <string.h>
+#include <float.h>
 
 #define N_(String) String
 #if ENABLE_NLS
@@ -44,12 +78,17 @@ static const char* unit_names[] = {
 	"compact",
 	"cyl",
 	"chs",
-	"%"
+	"%",
+	"kiB",
+	"MiB",
+	"GiB",
+	"TiB"
 };
 
 
 /**
- * Set the default unit used by subsequent calls to the PedUnit API.
+ * \brief Set the default \p unit used by subsequent calls to the PedUnit API.
+ *
  * In particular, this affects how locations inside error messages
  * (exceptions) are displayed.
  */
@@ -61,7 +100,7 @@ ped_unit_set_default (PedUnit unit)
 
 
 /**
- * Get the current default unit.
+ * \brief Get the current default unit.
  */
 PedUnit
 ped_unit_get_default ()
@@ -70,7 +109,7 @@ ped_unit_get_default ()
 }
 
 /**
- * Returns the byte size of a given unit.
+ * Get the byte size of a given \p unit.
  */
 long long
 ped_unit_get_size (PedDevice* dev, PedUnit unit)
@@ -78,17 +117,21 @@ ped_unit_get_size (PedDevice* dev, PedUnit unit)
 	PedSector cyl_size = dev->bios_geom.heads * dev->bios_geom.sectors;
 
 	switch (unit) {
-		case PED_UNIT_SECTOR:	return PED_SECTOR_SIZE;
+		case PED_UNIT_SECTOR:	return dev->sector_size;
 		case PED_UNIT_BYTE:	return 1;
 		case PED_UNIT_KILOBYTE:	return PED_KILOBYTE_SIZE;
 		case PED_UNIT_MEGABYTE:	return PED_MEGABYTE_SIZE;
 		case PED_UNIT_GIGABYTE:	return PED_GIGABYTE_SIZE;
 		case PED_UNIT_TERABYTE:	return PED_TERABYTE_SIZE;
-		case PED_UNIT_CYLINDER:	return cyl_size * PED_SECTOR_SIZE;
-		case PED_UNIT_CHS:	return PED_SECTOR_SIZE;
+		case PED_UNIT_KIBIBYTE:	return PED_KIBIBYTE_SIZE;
+		case PED_UNIT_MEBIBYTE:	return PED_MEBIBYTE_SIZE;
+		case PED_UNIT_GIBIBYTE:	return PED_GIBIBYTE_SIZE;
+		case PED_UNIT_TEBIBYTE:	return PED_TEBIBYTE_SIZE;
+		case PED_UNIT_CYLINDER:	return cyl_size * dev->sector_size;
+		case PED_UNIT_CHS:	return dev->sector_size;
 
 		case PED_UNIT_PERCENT:
-			return dev->length * PED_SECTOR_SIZE / 100;
+			return dev->length * dev->sector_size / 100;
 
 		case PED_UNIT_COMPACT:
 			ped_exception_throw (
@@ -104,7 +147,9 @@ ped_unit_get_size (PedDevice* dev, PedUnit unit)
 }
 
 /**
- * Returns a textual (non-internationalized) representation of a unit.
+ * Get a textual (non-internationalized) representation of a \p unit.
+ * 
+ * For example, the textual representation of PED_UNIT_SECTOR is "s".
  */
 const char*
 ped_unit_get_name (PedUnit unit)
@@ -112,8 +157,9 @@ ped_unit_get_name (PedUnit unit)
 	return unit_names[unit];
 }
 
-/*
- * Returns a unit based on its textual representation.
+/**
+ * Get a unit based on its textual representation: \p unit_name.
+ *
  * For example, ped_unit_get_by_name("Mb") returns PED_UNIT_MEGABYTE.
  */
 PedUnit
@@ -127,19 +173,6 @@ ped_unit_get_by_name (const char* unit_name)
 	return -1;
 }
 
-/**
- * Returns a string that describes the location "sector" on device "dev".
- * The string is described with the default unit, which is set
- * by ped_unit_set_default().
- * The returned string must be freed with ped_free().
- */
-char*
-ped_unit_format (PedDevice* dev, PedSector sector)
-{
-	PED_ASSERT (dev != NULL, return NULL);
-	return ped_unit_format_custom (dev, sector, default_unit);
-}
-
 static char*
 ped_strdup (const char *str)
 {
@@ -152,15 +185,21 @@ ped_strdup (const char *str)
 }
 
 /**
- * Returns a string that describes the location "sector" on device "dev".
- * The string is described with the desired unit.
+ * \brief Get a string that describes the location of the \p byte on device \p dev.
+ * 
+ * The string is described with the desired \p unit.
  * The returned string must be freed with ped_free().
  */
 char*
-ped_unit_format_custom (PedDevice* dev, PedSector sector, PedUnit unit)
+ped_unit_format_custom_byte (PedDevice* dev, PedSector byte, PedUnit unit)
 {
 	char buf[100];
+	PedSector sector = byte / dev->sector_size;
+	double d, w;
+	int p;
 
+	PED_ASSERT (dev != NULL, return NULL);
+	
 	/* CHS has a special comma-separated format. */
 	if (unit == PED_UNIT_CHS) {
 		PedCHSGeometry *chs = &dev->bios_geom;
@@ -171,38 +210,101 @@ ped_unit_format_custom (PedDevice* dev, PedSector sector, PedUnit unit)
 		return ped_strdup (buf);
 	}
 
-	/* Cylinders should be rounded down. */
-	if (unit == PED_UNIT_CYLINDER) {
-		snprintf (buf, 100, "%lldcyl",
-			  sector * PED_SECTOR_SIZE
-				  / ped_unit_get_size (dev, unit));
+	/* Cylinders, sectors and bytes should be rounded down... */
+	if (unit == PED_UNIT_CYLINDER
+	    || unit == PED_UNIT_SECTOR
+	    || unit == PED_UNIT_BYTE) {
+		snprintf (buf, 100, "%lld%s",
+			  byte / ped_unit_get_size (dev, unit),
+			  ped_unit_get_name (unit));
 		return ped_strdup (buf);
 	}
-
+	
 	if (unit == PED_UNIT_COMPACT) {
-		if (sector >= 10LL * PED_TERABYTE_SIZE / PED_SECTOR_SIZE)
+		if (byte >= 10LL * PED_TERABYTE_SIZE)
 			unit = PED_UNIT_TERABYTE;
-		else if (sector >= 10LL * PED_GIGABYTE_SIZE / PED_SECTOR_SIZE)
+		else if (byte >= 10LL * PED_GIGABYTE_SIZE)
 			unit = PED_UNIT_GIGABYTE;
-		else if (sector >= 10LL * PED_MEGABYTE_SIZE / PED_SECTOR_SIZE)
+		else if (byte >= 10LL * PED_MEGABYTE_SIZE)
 			unit = PED_UNIT_MEGABYTE;
 		else
 			unit = PED_UNIT_KILOBYTE;
 	}
 
-	snprintf (buf, 100, "%lld%s",
-		  ped_div_round_to_nearest (
-			  sector * PED_SECTOR_SIZE,
-			  ped_unit_get_size (dev, unit)),
-		  ped_unit_get_name (unit));
+	/* IEEE754 says that 100.5 has to be rounded to 100 (by printf) */
+	/* but 101.5 has to be rounded to 102... so we multiply by 1+E. */
+	/* This just divide by 2 the natural IEEE754 extended precision */
+	/* and won't cause any trouble before 1000 TB */
+	d = ((double)byte / (double)ped_unit_get_size (dev, unit))
+	    * (1. + DBL_EPSILON);
+	w = d + ( (d < 10. ) ? 0.005 :
+		  (d < 100.) ? 0.05  :
+			       0.5  );
+	p = (w < 10. ) ? 2 :
+	    (w < 100.) ? 1 :
+			 0 ;
+
+#ifdef __BEOS__
+	snprintf (buf, 100, "%.*f%s", p, d, ped_unit_get_name(unit));
+#else
+	snprintf (buf, 100, "%1$.*2$f%3$s", d, p, ped_unit_get_name (unit));
+#endif
+
 	return ped_strdup (buf);
 }
 
 /**
- * If str contains a valid description of a location on dev, then *sector
- * is modified to describe the location.  If no units are specified, then
- * the default unit is assumed.  This function returns 1 if str is
- * a valid location description, 0 otherwise.
+ * \brief Get a string that describes the location of the \p byte on device \p dev.
+ * 
+ * The string is described with the default unit, which is set
+ * by ped_unit_set_default().
+ * The returned string must be freed with ped_free().
+ */
+char*
+ped_unit_format_byte (PedDevice* dev, PedSector byte)
+{
+	PED_ASSERT (dev != NULL, return NULL);
+	return ped_unit_format_custom_byte (dev, byte, default_unit);
+}
+
+/**
+ * \brief Get a string that describes the location \p sector on device \p dev.
+ * 
+ * The string is described with the desired \p unit.
+ * The returned string must be freed with ped_free().
+ */
+char*
+ped_unit_format_custom (PedDevice* dev, PedSector sector, PedUnit unit)
+{
+	PED_ASSERT (dev != NULL, return NULL);
+	return ped_unit_format_custom_byte(dev, sector*dev->sector_size, unit);
+}
+
+/**
+ * \brief Get a string that describes the location \p sector on device \p dev.
+ * 
+ * The string is described with the default unit, which is set
+ * by ped_unit_set_default().
+ * The returned string must be freed with ped_free().
+ */
+char*
+ped_unit_format (PedDevice* dev, PedSector sector)
+{
+	PED_ASSERT (dev != NULL, return NULL);
+	return ped_unit_format_custom_byte (dev, sector * dev->sector_size,
+					    default_unit);
+}
+
+/**
+ * If \p str contains a valid description of a location on \p dev, 
+ * then \p *sector is modified to describe the location and a geometry 
+ * is created in \p *range describing a 2 units large area centered on 
+ * \p *sector.  If the \p range as described here would be partially outside 
+ * the device \p dev, the geometry returned is the intersection between the 
+ * former and the whole	device geometry.  If no units are specified, then the 
+ * default unit is assumed.  
+ *
+ * \return \c 1 if \p str is a valid location description, \c 0 otherwise
  */
 int
 ped_unit_parse (const char* str, PedDevice* dev, PedSector *sector,
@@ -309,15 +411,17 @@ parse_chs (const char* str, PedDevice* dev, PedSector* sector,
 				str, dev->path);
 		goto error_free_copy;
 	}
-	*range = ped_geometry_new (dev, *sector, 1);
+	if (range)
+		*range = ped_geometry_new (dev, *sector, 1);
 	ped_free (copy);
-	return 1;
+	return !range || *range != NULL;
 
 error_free_copy:
 	ped_free (copy);
 error:
 	*sector = 0;
-	*range = NULL;
+	if (range)
+		*range = NULL;
 	return 0;
 }
 
@@ -344,7 +448,14 @@ geometry_from_centre_radius (PedDevice* dev, PedSector sector, PedSector radius)
 static PedUnit
 parse_unit_suffix (const char* suffix, PedUnit suggested_unit)
 {
-	if (strlen (suffix) != 0) {
+	if (strlen (suffix) > 1 && tolower (suffix[1]) == 'i') {
+		switch (tolower (suffix[0])) {
+			case 'k': return PED_UNIT_KIBIBYTE;
+			case 'm': return PED_UNIT_MEBIBYTE;
+			case 'g': return PED_UNIT_GIBIBYTE;
+			case 't': return PED_UNIT_TEBIBYTE;
+		}
+	} else if (strlen (suffix) > 0) {
 		switch (tolower (suffix[0])) {
 			case 's': return PED_UNIT_SECTOR;
 			case 'b': return PED_UNIT_BYTE;
@@ -368,10 +479,19 @@ parse_unit_suffix (const char* suffix, PedUnit suggested_unit)
 }
 
 /**
- * If str contains a valid description of a location on dev, then *sector
- * is modified to describe the location.  If no units are specified, then
- * the desired unit is assumed.  This function returns 1 if str is
- * a valid location description, 0 otherwise.
+ * If \p str contains a valid description of a location on \p dev, then 
+ * \p *sector is modified to describe the location and a geometry is created
+ * in \p *range describing a 2 units large area centered on \p *sector.  If the
+ * \p range as described here would be partially outside the device \p dev, the
+ * geometry returned is the intersection between the former and the whole
+ * device geometry.  If no units are specified, then the default unit is
+ * assumed. 
+ *
+ * \throws PED_EXCEPTION_ERROR if \p str contains invalid description of a location
+ * \throws PED_EXCEPTION_ERROR if location described by \p str 
+ * 	is outside of the device \p dev->path
+ *
+ * \return \c 1 if \p str is a valid location description, \c 0 otherwise.
  */
 int
 ped_unit_parse_custom (const char* str, PedDevice* dev, PedUnit unit,
@@ -404,34 +524,38 @@ ped_unit_parse_custom (const char* str, PedDevice* dev, PedUnit unit,
 	}
 
 	unit_size = ped_unit_get_size (dev, unit);
-	radius = ped_div_round_up (unit_size, PED_SECTOR_SIZE) - 1;
+	radius = ped_div_round_up (unit_size, dev->sector_size) - 1;
 	if (radius < 0)
 		radius = 0;
 
-	*sector = num * unit_size / PED_SECTOR_SIZE;
+	*sector = num * unit_size / dev->sector_size;
 	/* negative numbers count from the end */
 	if (copy[0] == '-')
 		*sector += dev->length;
-	*range = geometry_from_centre_radius (dev, *sector, radius);
-	if (!*range) {
-		ped_exception_throw (
-			PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
-			_("The location %s is outside of the device %s."),
-			str, dev->path);
-		goto error_free_copy;
+	if (range) {
+		*range = geometry_from_centre_radius (dev, *sector, radius);
+		if (!*range) {
+			ped_exception_throw (
+				PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
+				_("The location %s is outside of the "
+				  "device %s."),
+				str, dev->path);
+			goto error_free_copy;
+		}
 	}
 	*sector = clip (dev, *sector);
 
 	ped_free (copy);
 	return 1;
 
-error_destroy_range:
-	ped_geometry_destroy (*range);
 error_free_copy:
 	ped_free (copy);
 error:
 	*sector = 0;
-	*range = NULL;
+	if (range)
+		*range = NULL;
 	return 0;
 }
 
+
+/** @} */
